from logging import getLogger
from smtplib import SMTP, SMTPNotSupportedError, SMTPServerDisconnected
from socket import error as SocketError
from socket import gethostname
from typing import Optional, Tuple

from dns.exception import Timeout
from dns.rdatatype import MX as rdtype_mx
from dns.rdtypes.ANY.MX import MX
from dns.resolver import (
    NXDOMAIN, YXDOMAIN, Answer, NoAnswer, NoNameservers, query)

from .constants import HOST_REGEX
from .email_address import EmailAddress
from .exceptions import (
    AddressNotDeliverableError, DNSConfigurationError, DNSTimeoutError,
    DomainNotFoundError, NoMXError, NoNameserverError, NoValidMXError)

LOGGER = getLogger(name=__name__)


class _ProtocolError(Exception):
    """
    Raised when there is an error during the SMTP conversation.
    Used only internally.
    """

    def __init__(self, command: str, code: int, message: bytes):
        self.command = command
        self.code = code
        self.message = message.decode(errors='ignore')

    def __str__(self):
        return f'{self.code} {self.message} (in reply to {self.command})'


def _get_mx_records(domain: str, timeout: int) -> list:
    """
    Return a list of hostnames in the MX record, raise an exception on
    any issues.
    """
    try:
        records = query(
            qname=domain, rdtype=rdtype_mx, lifetime=timeout)  # type: Answer
    except NXDOMAIN:
        raise DomainNotFoundError
    except NoNameservers:
        raise NoNameserverError
    except Timeout:
        raise DNSTimeoutError
    except YXDOMAIN:
        raise DNSConfigurationError
    except NoAnswer:
        raise NoMXError
    to_check = dict()
    for record in records:  # type: MX
        dns_str = record.exchange.to_text()  # type: str
        to_check[dns_str] = dns_str[:-1] if dns_str.endswith('.') else dns_str
    result = [k for k, v in to_check.items() if HOST_REGEX.search(string=v)]
    if not result:
        raise NoValidMXError
    return result


def _smtp_ehlo_tls(smtp: SMTP, helo_host: str):
    """
    Try and start the TLS session, fall back to unencrypted when
    unavailable.
    """
    code, message = smtp.ehlo(name=helo_host)
    if code >= 300:
        # EHLO bails out, no further SMTP commands are acceptable
        raise _ProtocolError('EHLO', code, message)
    try:
        smtp.starttls()
        code, message = smtp.ehlo(name=helo_host)
    except SMTPNotSupportedError:
        # The server does not support the STARTTLS extension
        pass
    except RuntimeError:
        # SSL/TLS support is not available to your Python interpreter
        pass


def _smtp_mail(smtp: SMTP, from_address: EmailAddress):
    'Send and evaluate the `MAIL FROM` command.'
    code, message = smtp.mail(sender=from_address.ace)
    if code >= 300:
        # MAIL FROM bails out, no further SMTP commands are acceptable
        raise _ProtocolError('MAIL FROM', code, message)


def _smtp_converse(
    mx_record: str, smtp_timeout: int, debug: bool, helo_host: str,
    from_address: EmailAddress, email_address: EmailAddress
):
    """
    Do the `SMTP` conversation, handle errors in the caller.

    Raise `_ProtocolError` on error, and `StopIteration` if the
    conversation points out an existing email.
    """
    if debug:
        LOGGER.debug(msg=f'Trying {mx_record} ...')
    with SMTP(timeout=smtp_timeout) as smtp:
        smtp.set_debuglevel(debuglevel=2 if debug else False)
        code, message = smtp.connect(host=mx_record)
        if code >= 400:
            raise _ProtocolError('connect', code, message)
        _smtp_ehlo_tls(smtp=smtp, helo_host=helo_host)
        _smtp_mail(smtp=smtp, from_address=from_address)
        code, message = smtp.rcpt(recip=email_address.ace)
        if code == 250:
            # Address valid, early exit
            raise StopIteration
        elif code >= 500:
            raise _ProtocolError('RCPT TO', code, message)


def _check_one_mx(
        error_messages: list, mx_record: str, helo_host: str,
        from_address: EmailAddress, email_address: EmailAddress,
        smtp_timeout: int, debug: bool) -> bool:
    """
    Check one MX server, return the `is_ambigious` boolean or raise
    `StopIteration` if this MX accepts the email.
    """
    try:
        _smtp_converse(
            mx_record=mx_record, smtp_timeout=smtp_timeout, debug=debug,
            helo_host=helo_host, from_address=from_address,
            email_address=email_address)
    except SMTPServerDisconnected:
        return True
    except (SocketError, _ProtocolError) as error:
        error_messages.append(f'{mx_record}: {error}')
        return False
    return True


def _check_mx_records(
        mx_records: list, smtp_timeout: int, helo_host: str,
        from_address: EmailAddress, email_address: EmailAddress,
        debug: bool) -> Optional[bool]:
    'Check the mx records for a given email address.'
    # TODO: Raise an ambigious exception, containing the messages? Will
    # be a breaking change.
    error_messages = []
    found_ambigious = False
    for mx_record in mx_records:
        try:
            found_ambigious |= _check_one_mx(
                error_messages=error_messages, mx_record=mx_record.rstrip('.'),
                helo_host=helo_host, from_address=from_address,
                email_address=email_address, smtp_timeout=smtp_timeout,
                debug=debug)
        except StopIteration:
            # Address valid, early exit
            return True
    # If any of the mx servers behaved ambigious, return None, otherwise raise
    # an exception containing the collected error messages.
    if not found_ambigious:
        raise AddressNotDeliverableError(error_messages=error_messages)


def mx_check(
    email_address: EmailAddress, debug: bool,
    from_address: Optional[EmailAddress] = None,
    helo_host: Optional[str] = None, smtp_timeout: int = 10,
    dns_timeout: int = 10
) -> Optional[bool]:
    """
    Return `True` if the host responds with a deliverable response code,
    `False` if not-deliverable. Also, return `None` if there if couldn't
    provide a conclusive result (e.g. temporary errors or graylisting).
    """
    host = helo_host or gethostname()
    from_address = from_address or email_address
    if email_address.domain_literal_ip:
        mx_records = [email_address.domain_literal_ip]
    else:
        mx_records = _get_mx_records(
            domain=email_address.domain, timeout=dns_timeout)
    return _check_mx_records(
        mx_records=mx_records, smtp_timeout=smtp_timeout, helo_host=host,
        from_address=from_address, email_address=email_address, debug=debug)



from typing import Iterable


class EmailValidationError(Exception):
    'Base class for all exceptions indicating validation failure.'
    message = 'Unknown error.'

    def __str__(self):
        return self.message


class AddressFormatError(EmailValidationError):
    'Raised when the email address has an invalid format.'
    message = 'Invalid email address.'


class FromAddressFormatError(EmailValidationError):
    """
    Raised when the from email address used for the MX check has an
    invalid format.
    """
    message = 'Invalid "From:" email address.'


class DomainBlacklistedError(EmailValidationError):
    """
    Raised when the domain of the email address is blacklisted on
    https://github.com/martenson/disposable-email-domains.
    """
    message = 'Domain blacklisted.'


class DomainNotFoundError(EmailValidationError):
    'Raised when the domain is not found.'
    message = 'Domain not found.'


class NoNameserverError(EmailValidationError):
    'Raised when the domain does not resolve by nameservers in time.'
    message = 'No nameserver found for domain.'


class DNSTimeoutError(EmailValidationError):
    'Raised when the domain lookup times out.'
    message = 'Domain lookup timed out.'


class DNSConfigurationError(EmailValidationError):
    """
    Raised when the DNS entries for this domain are falsely configured.
    """
    message = 'Misconfigurated DNS entries for domain.'


class NoMXError(EmailValidationError):
    'Raised then the domain has no MX records configured.'
    message = 'No MX record for domain found.'


class NoValidMXError(EmailValidationError):
    message = 'No valid MX record for domain found.'


class AddressNotDeliverableError(EmailValidationError):
    'Raised when a non-ambigious resulted lookup fails.'
    message = 'Email address undeliverable:'

    def __init__(self, error_messages: Iterable):
        self.error_messages = error_messages

    def __str__(self) -> str:
        return '\n'.join([self.message] + self.error_messages)



from logging import getLogger
from typing import Optional

from filelock import FileLock

from .email_address import EmailAddress
from .exceptions import DomainBlacklistedError
from .updater import (
    BLACKLIST_FILEPATH_INSTALLED, BLACKLIST_FILEPATH_TMP, LOCK_PATH,
    update_builtin_blacklist)

SetOrNone = Optional[set]
LOGGER = getLogger(__name__)


class DomainListValidator(object):
    'Check the provided email against domain lists.'
    domain_whitelist = set()
    domain_blacklist = set('localhost')
    _is_builtin_bl_used: bool = False

    def __init__(
            self, whitelist: SetOrNone = None, blacklist: SetOrNone = None):
        if whitelist:
            self.domain_whitelist = set(x.lower() for x in whitelist)
        if blacklist:
            self.domain_blacklist = set(x.lower() for x in blacklist)
        else:
            self._is_builtin_bl_used = True
            self.reload_builtin_blacklist()

    @property
    def _blacklist_path(self) -> str:
        'Return the path of the `blacklist.txt` that should be loaded.'
        try:
            # Zero size: file is touched to indicate the preinstalled
            # file is still fresh enough
            return BLACKLIST_FILEPATH_INSTALLED \
                if BLACKLIST_FILEPATH_TMP.stat().st_size == 0 \
                else BLACKLIST_FILEPATH_TMP
        except FileNotFoundError:
            return BLACKLIST_FILEPATH_INSTALLED

    def reload_builtin_blacklist(self):
        '(Re)load our built-in blacklist.'
        if not self._is_builtin_bl_used:
            return
        with FileLock(lock_file=LOCK_PATH):
            bl_path = self._blacklist_path
            LOGGER.debug(msg=f'(Re)loading blacklist from {bl_path}')
            try:
                with open(bl_path) as fd:
                    lines = fd.readlines()
            except FileNotFoundError:
                return
        self.domain_blacklist = set(
            x.strip().lower() for x in lines if x.strip())

    def __call__(self, address: EmailAddress) -> bool:
        'Do the checking here.'
        if address.domain in self.domain_whitelist:
            return True
        if address.domain in self.domain_blacklist:
            raise DomainBlacklistedError
        return True


domainlist_check = DomainListValidator()
# Start an optional update on module load
update_builtin_blacklist(
    force=False, background=True,
    callback=domainlist_check.reload_builtin_blacklist)



from typing import Optional

from idna.core import IDNAError, encode

from .exceptions import AddressFormatError


class EmailAddress(object):
    """
    Internally used class to hold an email address.

    This class featuers splitting the email address into user and domain
    part as well as converting internationalized domain name into the
    ASCII-compatible encoding (ACE) according to the IDNA standard.
    """

    def __init__(self, address: str):
        self._address = address

        # Split email address into user and domain part.
        try:
            self._user, self._domain = self._address.rsplit('@', 1)
        except ValueError:
            raise AddressFormatError

        # Convert internationalized domain name into the ACE encoding
        if self.domain_literal_ip:
            self._ace_domain = self._domain
        else:
            try:
                self._ace_domain = encode(self._domain).decode('ascii')
            except IDNAError:
                raise AddressFormatError

    @property
    def user(self) -> str:
        """
        The username part of the email address, that is the part before
        the "@" sign.
        """
        return self._user

    @property
    def domain(self) -> str:
        """
        The domain part of the email address, that is the part after the
        "@" sign.
        """
        return self._domain

    @property
    def domain_literal_ip(self) -> Optional[str]:
        """
        If the domain part of the email address is a literal IP address
        enclosed in brackets, that IP address (without the brackets) is
        returned. Otherwise, `None` is returned.
        """
        if self._domain.startswith('[') and self._domain.endswith(']'):
            return self._domain[1:-1]

    @property
    def ace(self) -> str:
        'The ASCII-compatible encoding for the email address.'
        return '@'.join((self._user, self._ace_domain))

    @property
    def ace_domain(self) -> str:
        """
        The ASCII-compatible encoding for the domain part of the email
        address.
        """
        return self._ace_domain



from ipaddress import IPv4Address, IPv6Address

from .constants import HOST_REGEX, LITERAL_REGEX, USER_REGEX
from .email_address import EmailAddress
from .exceptions import AddressFormatError


def _validate_ipv4_address(value: str):
    try:
        IPv4Address(address=value)
    except ValueError:
        return False
    return True


def _validate_ipv6_address(value: str) -> bool:
    """
    Return whether or not the `ip_str` string is a valid IPv6 address.
    """
    try:
        IPv6Address(address=value)
    except ValueError:
        return False
    return True


def _validate_ipv46_address(value: str) -> bool:
    return _validate_ipv4_address(value) or _validate_ipv6_address(value)


def regex_check(address: EmailAddress) -> bool:
    'Slightly adjusted email regex checker from the Django project.'

    # Validate user part.
    if not USER_REGEX.match(address.user):
        raise AddressFormatError

    # Validate domain part.
    if address.domain_literal_ip:
        literal_match = LITERAL_REGEX.match(address.ace_domain)
        if literal_match is None:
            raise AddressFormatError
        if not _validate_ipv46_address(literal_match[1]):
            raise AddressFormatError
    else:
        if HOST_REGEX.match(address.ace_domain) is None:
            raise AddressFormatError

    # All validations successful.
    return True



from http.client import HTTPResponse
from logging import getLogger
from pathlib import Path
from tempfile import gettempdir, gettempprefix
from threading import Thread
from time import time
from typing import Callable, Optional
from urllib.error import HTTPError
from urllib.request import Request, urlopen

try:
    from os import geteuid
except ImportError:
    def geteuid():
        'Windows does not have `os.geteuid()`.'
        return '1'


LOGGER = getLogger(__name__)
TMP_PATH = Path(gettempdir()).joinpath(
    f'{gettempprefix()}-py3-validate-email-{geteuid()}')
TMP_PATH.mkdir(exist_ok=True)
BLACKLIST_URL = (
    'https://raw.githubusercontent.com/martenson/disposable-email-domains/'
    'master/disposable_email_blocklist.conf')
LIB_PATH_DEFAULT = Path(__file__).resolve().parent.joinpath('data')
BLACKLIST_FILEPATH_INSTALLED = LIB_PATH_DEFAULT.joinpath('blacklist.txt')
BLACKLIST_FILEPATH_TMP = TMP_PATH.joinpath('blacklist.txt')
ETAG_FILEPATH_INSTALLED = LIB_PATH_DEFAULT.joinpath('blacklist.etag.txt')
ETAG_FILEPATH_TMP = TMP_PATH.joinpath('blacklist.etag.txt')
LOCK_PATH = TMP_PATH.joinpath('blacklistupdater.lock')


class BlacklistUpdater(object):
    """
    Optionally auto-update the built-in `blacklist.txt`, while using
    a temporary place to put the newly downloaded one to avoid read-only
    filesystem errors. If the installed `blacklist.txt` is fresh enough
    don't look for newer versions.
    """

    _refresh_when_older_than: int = 5 * 24 * 60 * 60  # 5 days

    def _read_etag(self) -> Optional[str]:
        'Read the etag header from the stored etag file when exists.'
        for path in [ETAG_FILEPATH_TMP, ETAG_FILEPATH_INSTALLED]:
            try:
                return path.read_text().strip()
            except FileNotFoundError:
                pass

    @property
    def _is_old(self) -> bool:
        'Return `True` if the locally stored file is old.'
        true_when_older_than = time() - self._refresh_when_older_than
        for path in [BLACKLIST_FILEPATH_TMP, BLACKLIST_FILEPATH_INSTALLED]:
            try:
                return path.stat().st_ctime < true_when_older_than
            except FileNotFoundError:
                pass
        return True  # no file found at all

    def _get_headers(self, force_update: bool = False) -> dict:
        'Compile a header with etag if available.'
        headers = dict()
        if force_update:
            return headers
        etag = self._read_etag()
        if not etag:
            return headers
        headers['If-None-Match'] = etag
        return headers

    def _download(self, headers: dict, blacklist_path: Path, etag_path: Path):
        'Downlad and store blacklist file.'
        LOGGER.debug(msg=f'Checking {BLACKLIST_URL}')
        request = Request(url=BLACKLIST_URL, headers=headers)
        response = urlopen(url=request)  # type: HTTPResponse
        # New data available
        LOGGER.debug(msg=f'Writing response into {blacklist_path}')
        blacklist_path.write_bytes(response.fp.read())
        if 'ETag' in response.headers:
            LOGGER.debug(msg=f'Storing ETag response into {etag_path}.')
            etag_path.write_text(response.headers['ETag'])

    def _install(self):
        """
        Download and store the blacklist file and the matching etag file
        into the library path. This is executed from setup.py upon
        installation of the library. Don't call this in your
        application.
        """
        LIB_PATH_DEFAULT.mkdir(exist_ok=True)
        self._download(
                headers={}, blacklist_path=BLACKLIST_FILEPATH_INSTALLED,
                etag_path=ETAG_FILEPATH_INSTALLED)

    def _process(self, force: bool = False):
        'Start optionally updating the blacklist.txt file, while locked.'
        if not force and not self._is_old:
            LOGGER.debug(msg='Not updating because file is fresh enough.')
            return
        try:
            self._download(
                headers=self._get_headers(force_update=force),
                blacklist_path=BLACKLIST_FILEPATH_TMP,
                etag_path=ETAG_FILEPATH_TMP)
        except HTTPError as exc:
            if exc.code == 304:
                # Not modified, update date on the tmp file
                LOGGER.debug(msg='Local file is fresh enough (same ETag).')
                BLACKLIST_FILEPATH_TMP.touch()
                return
            raise

    def process(
            self, force: bool = False, callback: Optional[Callable] = None):
        'Start optionally updating the blacklist.txt file.'
        # Locking to avoid multi-process update on multi-process startup
        # Import filelock locally because this module is als used by setup.py
        from filelock import FileLock
        with FileLock(lock_file=LOCK_PATH):
            self._process(force=force)
        # Always execute callback because multiple processes can have
        # different versions of blacklists (one before, one after
        # updating)
        if callback:
            callback()


def update_builtin_blacklist(
        force: bool = False, background: bool = True,
        callback: Callable = None) -> Optional[Thread]:
    """
    Update and reload the built-in blacklist. Return the `Thread` used
    to do the background update, so it can be `join()`-ed.
    """
    LOGGER.debug(msg='Starting optional update of built-in blacklist.')
    blacklist_updater = BlacklistUpdater()
    kwargs = dict(force=force, callback=callback)
    if not background:
        blacklist_updater.process(**kwargs)
        return
    bl_thread = Thread(target=blacklist_updater.process, kwargs=kwargs)
    bl_thread.start()
    return bl_thread



from logging import getLogger
from typing import Optional

from .domainlist_check import domainlist_check
from .email_address import EmailAddress
from .exceptions import (
    AddressFormatError, EmailValidationError, FromAddressFormatError)
from .mx_check import mx_check
from .regex_check import regex_check

LOGGER = getLogger(name=__name__)


def validate_email_or_fail(
        email_address: str, check_regex: bool = True, check_mx: bool = True,
        from_address: Optional[str] = None, helo_host: Optional[str] = None,
        smtp_timeout: int = 10, dns_timeout: int = 10,
        use_blacklist: bool = True, debug: bool = False) -> Optional[bool]:
    """
    Return `True` if the email address validation is successful, `None` if the
    validation result is ambigious, and raise an exception if the validation
    fails.
    """
    email_address = EmailAddress(email_address)
    if from_address is not None:
        try:
            from_address = EmailAddress(from_address)
        except AddressFormatError:
            raise FromAddressFormatError

    if check_regex:
        regex_check(email_address)
    if use_blacklist:
        domainlist_check(email_address)
    if not check_mx:
        return True
    return mx_check(
        email_address=email_address, from_address=from_address,
        helo_host=helo_host, smtp_timeout=smtp_timeout,
        dns_timeout=dns_timeout, debug=debug)


def validate_email(
        email_address: str, *args, **kwargs):
    """
    Return `True` or `False` depending if the email address exists
    or/and can be delivered.

    Return `None` if the result is ambigious.
    """
    try:
        return validate_email_or_fail(email_address, *args, **kwargs)
    except EmailValidationError as error:
        message = f'Validation for {email_address!r} failed: {error}'
        if kwargs.get('debug'):
            LOGGER.warning(msg=message)
        return False



import sys
from distutils import log
from pathlib import Path

from setuptools import find_packages, setup
from setuptools.command.build_py import build_py
from setuptools.command.develop import develop


def run_initial_updater(path: Path):
    'Download an initial blacklist.txt on install time.'
    # Only import the updater module to avoid requiring all the dependencies
    # and auto-running the updater.
    sys.path.append(str(path.joinpath('validate_email')))
    orig_dont_write_bytecode = sys.dont_write_bytecode
    sys.dont_write_bytecode = True
    try:
        from updater import BLACKLIST_FILEPATH_INSTALLED, BlacklistUpdater
        log.info(f'downloading blacklist to {BLACKLIST_FILEPATH_INSTALLED}')
        BlacklistUpdater()._install()
    finally:
        sys.path = sys.path[:-1]
        sys.dont_write_bytecode = orig_dont_write_bytecode


class DevelopCommand(develop):
    """
    Adapted version of the 'develop' command.

    After finishing the usual build run, download the blacklist and
    store it into the source directory, because that is from where the
    library will run in a developer install.
    """

    def run(self):
        super().run()
        if not self.dry_run:
            run_initial_updater(Path(__file__).parent)


class BuildPyCommand(build_py):
    """
    Adapted version of the 'build_py' command.

    After finishing the usual build run, download the blacklist and
    store it into the build directory. A subsequent 'install' step will
    copy the full contents of the build directory to the install
    target, thus including the blacklist.
    """

    def run(self):
        super().run()
        if not self.dry_run:
            run_initial_updater(Path(self.build_lib))


setup(
    name='py3-validate-email',
    version='0.2.14',
    packages=find_packages(exclude=['tests']),
    install_requires=['dnspython~=2.0', 'idna~=2.10', 'filelock~=3.0'],
    author='László Károlyi',
    author_email='laszlo@karolyi.hu',
    description=(
        'Email validator with regex, blacklisted domains and SMTP checking.'),
    long_description=Path(__file__).parent.joinpath('README.rst').read_text(),
    long_description_content_type='text/x-rst',
    keywords='email validation verification mx verify',
    url='http://github.com/karolyi/py3-validate-email',
    cmdclass=dict(build_py=BuildPyCommand, develop=DevelopCommand),
    license='LGPL',
)
